body {
  padding-top: 2em;
}
.tp-main h1 {
  text-align: center;
  font-size: 4em;
  margin-bottom: 1em;
}
.tp-main .lead {
  font-size: 2em;
}
form {
  display: inline;
}
form input {
  vertical-align: middle;
}
.row {
  text-align: center;
}
.col-md-3 {
  margin-bottom: 1em;
}

input.shadow {
  -moz-box-shadow: 2px 2px 4px rgba(0,0,0,0.4);
  -webkit-box-shadow: 2px 2px 4px rgba(0,0,0,0.4);
  /* CSS3 */
  box-shadow: 2px 2px 4px rgba(0,0,0,0.4);
  /* For IE 8 */
  -ms-filter: progid:DXImageTransform.Microsoft.Shadow(Strength=2,Direction=135,Color='#666666');
  /* For IE 5.5 - 7 */
  filter: progid:DXImageTransform.Microsoft.Shadow(Strength=2,Direction=135,Color='#666666');
}

input.shadow:hover {
  -moz-box-shadow: 2px 2px 4px rgba(0,0,0,0.8);
  -webkit-box-shadow: 2px 2px 4px rgba(0,0,0,0.8);
  /* CSS3 */
  box-shadow: 2px 2px 4px rgba(0,0,0,0.8);
  /* For IE 8 */
  -ms-filter: progid:DXImageTransform.Microsoft.Shadow(Strength=2,Direction=135,Color='#222222');
  /* For IE 5.5 - 7 */
  filter: progid:DXImageTransform.Microsoft.Shadow(Strength=2,Direction=135,Color='#222222');
}
